import json

from flask import request
from werkzeug.exceptions import BadRequest


__all__ = [
    'Handler', 'ResponseHandler', 'RequestHandler',
    'JSONRequestMixin', 'JSONResponseMixin'
]


class Handler(object):

    def __init__(self, endpoint, payload_key='payload', **params):

        self.endpoint = endpoint
        self.params = params
        self.data = None
        self._errors = None
        self.payload_key = payload_key

    def handle(self, data, **kwargs):
        """Invoke the handler to process the provided data.  Concrete classes
        should override this method to provide specific capabilties for the
        choosen method of marshaling and serializing data.

        :param data: The data to be processed by the Handler.
        :returns: The data processed by the Handler
        :rtype: mixed

        Here's an example of a RequestHandler integrating with the Kim library.

        .. code-block:: python

            def handle(self, data):
                try:
                    if self.many:
                        return self.mapper.many(raw=self.raw,
                                                **self.mapper_kwargs) \
                            .marshal(data, role=self.role)
                    else:
                        return self.mapper(data=data, raw=self.raw,
                                           **self.mapper_kwargs) \
                            .marshal(role=self.role)
                except MappingInvalid as e:
                    self.errors = e.errors

        .. seealso:
            :meth:`Handler.process`
        """
        return data

    def process(self, data=None, **kwargs):
        """Process the provided data and invoke :meth:`Handler.handle` method for this
        Handler class.

        :params data: The data being processed.
        :returns: self
        :rtype: :class:`Handler`

        .. code-block:: python

            def post(self, *args, **kwargs):
                self.request = self.get_request_handler()
                self.request.process(self.get_data())
                return self.get_create_response()

        .. seealso:
            :meth:`Handler.process`
        """
        self.data = self.handle(data, **kwargs)
        return self


class JSONResponseMixin(object):
    """Provides handling for serializing the response data as a JSON string.
    """

    def get_response_data(self):
        """serialzie the response data and payload_key as a JSON string.

        :returns: JSON serialized string
        :rtype: bytes
        """

        return json.dumps({self.payload_key: self.data})


class JSONRequestMixin(object):
    """Provides handling for fetching JSON data from the FLask request object.
    """

    def get_request_data(self):
        """Pull JSON from the Flask request object.

        :returns: Deserialized JSON data.
        :rtype: mixed
        :raises: :class:`flask.exceptions.JSONBadRequest`
        """

        try:
            return request.json or {}
        except BadRequest:
            return self.endpoint.return_error(
                400,
                payload={'message': 'Invalid JSON data provided'}
            )


class RequestHandler(Handler, JSONRequestMixin):
    """Basic default RequestHandler that expects the will pull JSON from the Flask request
    object and return it.
    """

    def process(self, data=None):
        """Fetch incoming data from the Flask request object when no data is supplied
        to the process method.  By default, the RequestHandler expects the
        incoming data to be sent as JSON.
        """

        return super(RequestHandler, self).process(data=data or self.get_request_data())


class ResponseHandler(Handler, JSONResponseMixin):
    """Basic default ResponseHanlder that expects the data passed to it to be JSON
    serializable without any modifications.
    """
    pass





__all__ = ['ArrestedAPI']


class ArrestedAPI(object):
    """ArrestedAPI defines versions of your api on which :class:`.Endpoint` are registered
    It acts like ``Flasks``  Blueprint object with a few minor differences.
    """

    def __init__(self, app=None, url_prefix='', before_all_hooks=None,
                 after_all_hooks=None):
        """Constructor to create a new ArrestedAPI object.

        :param app: Flask app object.
        :param url_prefix: Specify a url prefix for all resources attached to this API.
            Typically this is used for specifying the API version.
        :param before_all_hooks: A list containing funcs which will be called before
            every request made to any resource registered on this Api.
        :param after_all_hooks: A list containing funcs which will be called after
            every request made to any resource registered on this Api.

        Usage::

            app = Flask(__name__, url_prefix='/v1')
            api_v1 = ArrestedAPI(app)

        """

        self.before_all_hooks = before_all_hooks or []
        self.after_all_hooks = after_all_hooks or []
        self.url_prefix = url_prefix
        self.deferred = []
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """Initialise the ArrestedAPI object by storing a pointer to a Flask app object.
        This method is typically used when initialisation is deferred.

        :param app: Flask application object

        Usage::

            app = Flask(__name__)
            ap1_v1 = ArrestedAPI()
            api_v1.init_app(app)
        """

        self.app = app
        if self.deferred:
            self.register_all(self.deferred)

    def register_resource(self, resource, defer=False):
        """Register a :class:`.Resource` blueprint object against the Flask app object.

        :param resource: :class:`.Resource` or :class:`flask.Blueprint`
            object.
        :param defer: Optionally specify that registering this resource should be
            deferred.  This option is useful when users are creating their
            Flask app instance via a factory.

        **Deferred resource registration**

        Resources can optionally be registered in a deferred manner.  Simply pass
        `defer=True` to :meth:`.ArrestedAPI.register_resource` to attach the resource to
        the API without calling register_blueprint.

        This is useful when you're using the factory pattern for creating your Flask app
        object as demonstrated below. Deferred resource will not be registered until the
        ArrestedAPI instance is initialised with the Flask app object.

        Usage::

            api_v1 = ArrestedAPI(prefix='/v1')
            characters_resource = Resource(
                'characters', __name__, url_prefix='/characters'
            )
            ap1_v1.register_resource(characters_resource, defer=True)

            def create_app():

                app = Flask(__name__)
                api_v1.init_app(app) # deferred resources are now registered.
        """
        if defer:
            self.deferred.append(resource)
        else:
            resource.init_api(self)
            self.app.register_blueprint(resource, url_prefix=self.url_prefix)

    def register_all(self, resources):
        """Register each resource from an iterable.

        :params resources: An iterable containing :class:`.Resource` objects

        Usage::

            characters_resource = Resource(
                'characters', __name__, url_prefix='/characters'
            )
            planets_resource = Resource('planets', __name__, url_prefix='/planets')
            api_v1 = ArrestedAPI(prefix='/v1')
            api_v1.register_all([characters_resource, planets_resource])
        """
        for resource in resources:
            self.register_resource(resource)




__all__ = [
    'GetListMixin', 'CreateMixin', 'GetObjectMixin', 'PutObjectMixin',
    'PatchObjectMixin', 'DeleteObjectMixin', 'ObjectMixin'
]


class HTTPMixin(object):
    """
    """

    def _response(self, body, status):
        """
        """

        return (self.make_response(body, status=status))


class GetListMixin(HTTPMixin):
    """Base ListMixin class that defines the expected API for all ListMixins
    """

    def get_objects(self):
        """
        """

        raise NotImplementedError()

    def list_response(self, status=200):
        """Pull the processed data from the response_handler and return a response.

        :param status: The HTTP status code returned with the response

        .. seealso:
            :meth:`Endpoint.make_response`
            :meth:`Endpoint.handle_get_request`
        """

        return self._response(self.response.get_response_data(), status=status)

    def handle_get_request(self):
        """Handle incoming GET request to an Endpoint and return an
        array of results by calling :meth:`.GetListMixin.get_objects`.

        .. seealso::
            :meth:`GetListMixin.get_objects`
            :meth:`Endpoint.get`
        """
        self.objects = self.get_objects()
        self.response = self.get_response_handler()

        self.response.process(self.objects)

        return self.list_response()


class CreateMixin(HTTPMixin):
    """Base CreateMixin class that defines the expected API for all CreateMixins
    """

    def save_object(self, obj):
        """Called by :meth:`CreateMixin.handle_post_request` ater the incoming data has
        been marshalled by the RequestHandler.

        :param obj: The marhsaled object from RequestHandler.
        """
        return obj

    def create_response(self, status=201):
        """Generate a Response object for a POST request.  By default, the newly created
        object will be passed to the specified ResponseHandler and will be serialized
        as the response body.
        """
        self.response = self.get_response_handler()
        self.response.process(self.obj)

        return self._response(self.response.get_response_data(), status=status)

    def handle_post_request(self):
        """Handle incoming POST request to an Endpoint and marshal the request data
        via the specified RequestHandler.  :meth:`.CreateMixin.save_object`. is then
        called and must be implemented by mixins implementing this interfce.

        .. seealso::
            :meth:`CreateMixin.save_object`
            :meth:`Endpoint.post`
        """
        self.request = self.get_request_handler()
        self.obj = self.request.process().data

        self.save_object(self.obj)
        return self.create_response()


class ObjectMixin(object):
    """Mixin that provides an interface for working with single data objects
    """

    allow_none = False

    def get_object(self):
        """Called by :meth:`GetObjectMixin.handle_get_request`.  Concrete classes should
        implement this method and return object typically by id.

        :raises: NotImplementedError
        """
        raise NotImplementedError()

    def object_response(self, status=200):
        """Generic response generation for Endpoints that return a single
        serialized object.

        :param status: The HTTP status code returned with the response
        :returns: Response object
        """

        self.response = self.get_response_handler()
        self.response.process(self.obj)
        return self._response(self.response.get_response_data(), status=status)

    @property
    def obj(self):
        """Returns the value of :meth:`ObjectMixin.get_object` and sets a private
        property called _obj.  This property ensures the logic around allow_none
        is enforced across Endpoints using the Object interface.

        :raises: :class:`werkzeug.exceptions.BadRequest`
        :returns: The result of :meth:ObjectMixin.get_object`
        """
        if not getattr(self, '_obj', None):
            self._obj = self.get_object()
            if self._obj is None and not self.allow_none:
                self.return_error(404)

        return self._obj

    @obj.setter
    def obj(self, value):
        """Sets the value of the private _obj property.
        """
        self._obj = value


class GetObjectMixin(HTTPMixin, ObjectMixin):
    """Base GetObjectMixins class that defines the expected API for all GetObjectMixins
    """

    def handle_get_request(self):
        """Handle incoming GET request to an Endpoint and return a
        single object by calling :meth:`.GetListMixin.get_object`.

        .. seealso::
            :meth:`GetListMixin.get_objects`
            :meth:`Endpoint.get`
        """
        return self.object_response()


class PutObjectMixin(HTTPMixin, ObjectMixin):
    """Base PutObjectMixins class that defines the expected API for all PutObjectMixin
    """

    def put_request_response(self, status=200):
        """Pull the processed data from the response_handler and return a response.

        :param status: The HTTP status code returned with the response

        .. seealso:
            :meth:`ObjectMixin.object_response`
            :meth:`Endpoint.handle_put_request`
        """

        return self.object_response(status=status)

    def handle_put_request(self):
        """
        """
        obj = self.obj
        self.request = self.get_request_handler()
        self.request.process()

        self.update_object(obj)
        return self.put_request_response()

    def update_object(self, obj):
        """Called by :meth:`PutObjectMixin.handle_put_request` ater the incoming data has
        been marshalled by the RequestHandler.

        :param obj: The marhsaled object from RequestHandler.
        """
        return obj


class PatchObjectMixin(HTTPMixin, ObjectMixin):
    """Base PatchObjectMixin class that defines the expected API for all PatchObjectMixin
    """

    def patch_request_response(self, status=200):
        """Pull the processed data from the response_handler and return a response.

        :param status: The HTTP status code returned with the response

        .. seealso:
            :meth:`ObjectMixin.object_response`
            :meth:`Endpoint.handle_put_request`
        """

        return self.object_response(status=status)

    def handle_patch_request(self):
        """
        """
        obj = self.obj
        self.request = self.get_request_handler()
        self.request.process()

        self.patch_object(obj)
        return self.patch_request_response()

    def patch_object(self, obj):
        """Called by :meth:`PatchObjectMixin.handle_patch_request` ater the
        incoming data has been marshalled by the RequestHandler.

        :param obj: The marhsaled object from RequestHandler.
        """
        return obj


class DeleteObjectMixin(HTTPMixin, ObjectMixin):
    """Base DeletehObjecttMixin class that defines the expected API for
    all DeletehObjecttMixins.
    """

    def delete_request_response(self, status=204):
        """Pull the processed data from the response_handler and return a response.

        :param status: The HTTP status code returned with the response

        """
        return (self.make_response('', status=status))

    def handle_delete_request(self):
        """
        """
        self.delete_object(self.obj)
        return self.delete_request_response()

    def delete_object(self, obj):
        """Called by :meth:`DeleteObjectMixin.handle_delete_request`.

        :param obj: The marhsaled object being deleted.
        """
        return obj



from flask import Blueprint


__all__ = ['Resource']


class Resource(Blueprint):
    """Resource extends Flasks existing Blueprint object and provides some utilty methods
    that make registering :class:`Endpoint` simpler.
    """

    def __init__(
            self, name, import_name, api=None, before_all_hooks=None,
            after_all_hooks=None, *args, **kwargs):
        """Construct a new Reosurce blueprint.  In addition to the normal Blueprint
        options, Resource accepts kwargs to set request middleware.

        :param api: API instance being the resource is being registered against.
        :param before_all_hooks: A list of middleware functions that will be applied
            before every request.
        :param after_all_hooks: A list of middleware functions that will be applied
            after every request.

        **Middleware**

        Arrested supports applying request middleware at every level of the application
        stack.  Resource middleware will be applied for every Endpoint registered on
        that Resource.

        .. code-block:: python

            character_resource = Resource(
                'characters', __name__,
                url_prefix='/characters',
                before_all_hooks=[log_request]
            )

        Arrested request middleware works differently from the way Flask middleware does.
        Middleware registered at the Resource and API level are consumed by the
        :meth:`arrested.Endpoint.dispatch_request` rather than being fired via the Flask
        app instance.  This is so we can pass the instance of the endpoint handling the
        request to each piece of middleware registered on the API or Resource.

        .. code-block:: python

            def api_key_required(endpoint):

                if request.args.get('api_key', None) is None:
                    endpoint.return_error(422)
                else:
                    get_client(request.args['api_key'])

            character_resource = Resource(
                'characters', __name__,
                url_prefix='/characters',
                before_all_hooks=[api_key_required]
            )

        Please note, Flask's normal App and Blueprint middleware can still be used
        as normal, it just doesn't recieve the instance of the endpoint registered to
        handle the request.

        """

        self.before_all_hooks = before_all_hooks or []
        self.after_all_hooks = after_all_hooks or []

        super(Resource, self).__init__(name, import_name, *args, **kwargs)
        if api is not None:
            self.init_api(api)

    def init_api(self, api):
        """Registered the instance of :class:`.ArrestedAPI` the :class:`.Resource`
        is being registered against.

        :param api: API instance being the resource is being registered against.
        """
        self.api = api

    def add_endpoint(self, endpoint):
        """Register an :class:`.Endpoint` aginst this resource.

        :param endpoint: :class:`.Endpoint` API Endpoint class

        Usage::

            foo_resource = Resource('example', __name__)
            class MyEndpoint(Endpoint):

                url = '/example'
                name = 'myendpoint'

            foo_resource.add_endpoint(MyEndpoint)
        """

        if self.url_prefix:
            url = '{prefix}{url}'.format(prefix=self.url_prefix, url=endpoint.url)
        else:
            url = endpoint.url

        self.add_url_rule(
            url,
            view_func=endpoint.as_view(endpoint.get_name()),
        )



import pytest

from mock import patch, Mock
from flask_sqlalchemy import SQLAlchemy

from arrested import Endpoint
from arrested.exceptions import ArrestedException
from arrested.contrib.sql_alchemy import (
    DBMixin,
    DBListMixin,
    DBCreateMixin,
    DBObjectMixin
)


class Model(object):

    id = 'foo'
    name = 'foo'


def test_get_db_session(app):

    db = SQLAlchemy(app)
    assert DBMixin().get_db_session() == db.session


def test_db_list_mixin_requires_get_query(app):

    mixin = DBListMixin()
    with pytest.raises(NotImplementedError):
        mixin.get_query()


def test_db_list_mixin_get_objects(app):


    with patch.object(DBListMixin, 'get_query') as query_mock:
        query_mock.return_value.all.return_value = ['foo', 'bar']
        res = DBListMixin().get_objects()
        assert res == ['foo', 'bar']


def test_db_list_mixin_get_result(app):

    query_mock = Mock()
    query_mock.return_value.all.return_value = ['foo', 'bar']

    res = DBListMixin().get_result(query_mock.return_value)
    assert res == ['foo', 'bar']
    query_mock.return_value.all.assert_called_once()


def test_db_create_mixin_save_object(app):

    session_mock = Mock()
    fake = Mock()

    with patch.object(
        DBCreateMixin, 'get_db_session', return_value=session_mock) as get_session_mock:

        DBCreateMixin().save_object(fake)

        get_session_mock.assert_called_once()

        session_mock.add.assert_called_once_with(fake)
        session_mock.commit.assert_called_once()


def test_db_object_mixin_get_object(app):

    mixin = DBObjectMixin()
    mixin.kwargs = {'obj_id': 1}  # Simulate kwargs being set by an Endpoint class.
    mixin.model = Model
    query_mock = patch.object(DBObjectMixin, 'get_query')
    mock_query = query_mock.start()
    mock_query.return_value.filter.return_value.one_or_none.return_value = 'foo'

    res = mixin.get_object()

    assert res == 'foo'

    query_mock.stop()


def test_db_object_mixin_get_result(app):

    mock_query = Mock()
    mock_query.return_value.one_or_none.return_value = 'foo'
    mixin = DBObjectMixin()
    mixin.kwargs = {'obj_id': 1}  # Simulate kwargs being set by an Endpoint class.

    res = mixin.get_result(mock_query.return_value)

    assert res == 'foo'
    mock_query.return_value.one_or_none.assert_called_once()


def test_db_object_mixin_filter_by_id_model_must_be_set(app):

    mock_query = Mock()
    mixin = DBObjectMixin()
    mixin.kwargs = {'obj_id': 1}  # Simulate kwargs being set by an Endpoint class.

    with pytest.raises(ArrestedException):
        mixin.filter_by_id(mock_query.return_value)


def test_db_object_mixin_filter_by_id_invalid_id_field(app):

    mock_query = Mock()
    mixin = DBObjectMixin()
    mixin.model = Model
    mixin.model_id_param = 'bar'
    mixin.kwargs = {'obj_id': 1}  # Simulate kwargs being set by an Endpoint class.

    with pytest.raises(ArrestedException):
        mixin.filter_by_id(mock_query.return_value)



def test_db_object_mixin_filter_by_id(app):

    mock_query = Mock()
    mixin = DBObjectMixin()
    mixin.model = Model
    mixin.kwargs = {'obj_id': 1}  # Simulate kwargs being set by an Endpoint class.

    mixin.filter_by_id(mock_query.return_value)

    params = (Model.id == 1)
    mock_query.return_value.filter.assert_called_once_with(params)


def test_db_update_mixin_update_object(app):

    session_mock = Mock()
    fake = Mock()

    with patch.object(
        DBObjectMixin, 'get_db_session', return_value=session_mock) as get_session_mock:

        DBObjectMixin().update_object(fake)

        get_session_mock.assert_called_once()

        session_mock.add.assert_called_once_with(fake)
        session_mock.commit.assert_called_once()


def test_db_delete_mixin_delete_object(app):

    session_mock = Mock()
    fake = Mock()

    with patch.object(
        DBObjectMixin, 'get_db_session', return_value=session_mock) as get_session_mock:

        DBObjectMixin().delete_object(fake)

        get_session_mock.assert_called_once()

        session_mock.delete.assert_called_once_with(fake)
        session_mock.commit.assert_called_once()


def test_db_mixins_mro(app):

    class IndexApi(Endpoint, DBListMixin, DBCreateMixin):
        pass


    class ObjectApi(Endpoint, DBObjectMixin):
        pass



import json
from werkzeug.wrappers import Response

from itertools import chain

from flask import Response, abort, request, current_app
from flask.views import MethodView

from .handlers import ResponseHandler, RequestHandler


class Endpoint(MethodView):
    """The Endpoint class represents the HTTP methods that can be called against an
    Endpoint inside of a particular resource.
    """

    #: list containing the permitted HTTP methods this endpoint accepts
    methods = ["GET", "POST", "PUT", "PATCH", "DELETE"]

    #: The name used to register this endpoint with Flask's url_map
    name = None

    #: A :class:`.ResponseHandler` class
    response_handler = ResponseHandler

    #: A :class:`.RequestHandler` class
    request_handler = RequestHandler

    #: The URL this endpoint is mapped against. This will build on top of any url_prefix
    #: defined at the API and Resource level
    url = ''

    #: A list of functions called before any specific request handler methods are called
    before_all_hooks = []

    #: A list of functions called before GET requests are dispatched
    before_get_hooks = []

    #: A list of functions called after GET requests are dispatched
    after_get_hooks = []

    #: A list of functions called before POST requests are dispatched
    before_post_hooks = []

    #: A list of functions called after POST requests are dispatched
    after_post_hooks = []

    #: A list of functions called before PUT requests are dispatched
    before_put_hooks = []

    #: A list of functions called after PUT requests are dispatched
    after_put_hooks = []

    #: A list of functions called before PATCH requests are dispatched
    before_patch_hooks = []

    #: A list of functions called after PATCH requests are dispatched
    after_patch_hooks = []

    #: A list of functions called before DELETE requests are dispatched
    before_delete_hooks = []

    #: A list of functions called after DELETE requests are dispatched
    after_delete_hooks = []

    #: A list of functions called after all requests are dispatched
    after_all_hooks = []

    def dispatch_request(self, *args, **kwargs):
        """Dispatch the incoming HTTP request to the appropriate handler.
        """
        self.args = args
        self.kwargs = kwargs
        meth = request.method.lower()

        if request.method not in self.methods:
            self.return_error(405)

        resource = current_app.blueprints.get(request.blueprint, None)
        if resource:
            for hook in chain(resource.api.before_all_hooks, resource.before_all_hooks):
                hook(self)

        before_hooks = getattr(self, 'before_{method}_hooks'.format(method=meth))
        for hook in before_hooks:
            hook(self)

        resp = super(Endpoint, self).dispatch_request(*args, **kwargs)
        resp = self.make_response(resp)

        after_hooks = getattr(self, 'after_{method}_hooks'.format(method=meth))
        for hook in after_hooks:
            resp = hook(self, resp)

        if resource:
            for hook in chain(resource.after_all_hooks, resource.api.after_all_hooks):
                resp = hook(self, resp)

        return resp

    @classmethod
    def get_name(cls):
        """Returns the user provided name or the lower() class name for use when
        registering the Endpoint with a :class:`Resource`.

        :returns: registration name for this Endpoint.
        :rtype: string
        """
        return cls.name or cls.__name__.lower()

    def make_response(self, rv, status=200, headers=None, mime='application/json'):
        """Create a response object using the :class:`flask.Response` class.

        :param rv: Response value.  If the value is not an instance
            of :class:`werkzeug.wrappers.Response` it will be converted
            into a Response object.
        :param status: specify the HTTP status code for this response.
        :param mime: Specify the mimetype for this request.
        :param headers: Specify dict of headers for the response.

        """
        if not isinstance(rv, Response):
            resp = Response(
                response=rv,
                headers=headers,
                mimetype=mime,
                status=status
            )
        else:
            resp = rv

        return resp

    def get_response_handler_params(self, **params):
        """Return a dictionary of options that are passed to the
        specified :class:`ResponseHandler`.

        :returns: Dictionary of :class:`ResponseHandler` config options.
        :rtype: dict
        """
        return params

    def get_response_handler(self):
        """Return the Endpoints defined :attr:`Endpoint.response_handler`.

        :returns: A instance of the Endpoint specified :class:`ResonseHandler`.
        :rtype: :class:`ResponseHandler`
        """
        assert self.response_handler is not None, \
            'Please define a response_handler ' \
            ' for Endpoint: %s' % self.__class__.__name__

        return self.response_handler(self, **self.get_response_handler_params())

    def get_request_handler_params(self, **params):
        """Return a dictionary of options that are passed to the
        specified :class:`RequestHandler`.

        :returns: Dictionary of :class:`RequestHandler` config options.
        :rtype: dict
        """
        return params

    def get_request_handler(self):
        """Return the Endpoints defined :attr:`Endpoint.request_handler`.

        :returns: A instance of the Endpoint specified :class:`RequestHandler`.
        :rtype: :class:`RequestHandler`
        """
        assert self.request_handler is not None, \
            'Please define a request_handler ' \
            ' for Endpoint: %s' % self.__class__.__name__

        return self.request_handler(self, **self.get_request_handler_params())

    def return_error(self, status, payload=None):
        """Error handler called by request handlers when an error occurs and the request
        should be aborted.

        Usage::

            def handle_post_request(self, *args, **kwargs):

                self.request_handler = self.get_request_handler()
                try:
                    self.request_handler.process(self.get_data())
                except SomeException as e:
                    self.return_error(400, payload=self.request_handler.errors)

                return self.return_create_response()
        """
        resp = None
        if payload is not None:
            payload = json.dumps(payload)
            resp = self.make_response(payload, status=status)

        if status in [405]:
            abort(status)
        else:
            abort(status, response=resp)

    def get(self, *args, **kwargs):
        """Handle Incoming GET requests and dispatch to handle_get_request method.
        """
        return self.handle_get_request()

    def post(self, *args, **kwargs):
        """Handle Incoming POST requests and dispatch to handle_post_request method.
        """
        return self.handle_post_request()

    def put(self, *args, **kwargs):
        """Handle Incoming PUT requests and dispatch to handle_put_request method.
        """
        return self.handle_put_request()

    def patch(self, *args, **kwargs):
        """Handle Incoming PATCH requests and dispatch to handle_patch_request method.
        """
        return self.handle_patch_request()

    def delete(self, *args, **kwargs):
        """Handle Incoming DELETE requests and dispatch to handle_delete_request method.
        """
        return self.handle_delete_request()



from flask import request
from kim.exception import MappingInvalid

from ..handlers import Handler, RequestHandler, ResponseHandler
from ..endpoint import Endpoint


__all__ = ['KimHandler', 'KimRequestHandler', 'KimResponseHandler', 'KimEndpoint']


class KimHandler(Handler):
    """Base handler providing Request/Response handling for the Kim serialization and
    marshaling framework.
    """

    def __init__(self, endpoint, *args, **params):
        """KimHandler constructor that creates a new instance of a :class:`KimHandler`.

        :param mapper_class: a Kim Mapper class
        :param mapper_kwargs: a dictionary of config options passed to the mapper each
            time it is instantiated.
        :param role: Optionally provide a role to be used by kim when serializing or
            marshaling.
        :param many: Pass the many option to Kim when marshaling or serializing.
        :param raw: Pass the raw option to Kim when marshaling or serializing.
        :returns: None
        """
        super(KimHandler, self).__init__(endpoint, *args, **params)

        self.mapper_class = params.pop('mapper_class', None)
        self.mapper_kwargs = params.pop('mapper_kwargs', {})
        self.role = params.pop('role', '__default__')
        self.many = params.pop('many', False)
        self.raw = params.pop('raw', False)
        self.obj = params.pop('obj', None)
        self.partial = params.pop('partial', False)

    @property
    def mapper(self):
        """Proxy to the provided mapper_class property.

        :raises: Exception when required mapper_class param is not provided
        :returns: Kim Mapper class
        """
        if self.mapper_class is None:
            raise Exception('KimHandler requires a Kim mapper_class')

        return self.mapper_class


class KimResponseHandler(KimHandler, ResponseHandler):
    """ResponseHandler for the Kim serialization and marshaling framework.

    .. code-block:: python

        from arrested.contrib.kim import KimResponseHandler

        class PlanetsEndpoint(Endpoint, DBListMixin):

            name = 'list'
            response_handler = KimResponseHandler

            def get_response_handler_params(self, **params):

                params = super(PlanetsEndpoint, self).get_response_handler_params(**params)
                params['mapper_class'] = PlanetMapper
                params['many'] = True

                return params

            def get_query(self):

                return db.session.query(Planet).order_by(
                    Planet.created_at.desc()
                )
    """

    def handle(self, data, **kwargs):
        """Run serialization for the specified mapper_class.

        Supports both .serialize and .many().serialize Kim interfaces.

        :param data: Objects to be serialized.
        :returns: Serialized data according to mapper configuration
        """
        if self.many:
            return self.mapper.many(raw=self.raw, **self.mapper_kwargs).serialize(
                data, role=self.role
            )
        else:
            return self.mapper(obj=data, raw=self.raw, **self.mapper_kwargs).serialize(
                role=self.role
            )


class KimRequestHandler(KimHandler, RequestHandler):
    """RequestHanlder for the Kim marshaling and serialization framework.

    .. code-block:: python

        from arrested.contrib.kim import KimRequestHandler

        class PlanetsEndpoint(Endpoint, DBListMixin):

            name = 'list'
            request_handler = KimRequestHandler

            def get_request_handler_params(self, **params):

                params = super(PlanetsEndpoint, self).get_response_handler_params(**params)
                params['mapper_class'] = PlanetMapper
                params['many'] = False

                return params

            def get_query(self):

                return db.session.query(Planet).order_by(
                    Planet.created_at.desc()
                )
    """

    def handle(self, data, error_status=422, **kwargs):
        """Run marshalling for the specified mapper_class.

        * TODO(mike) Allows users to control the error reponse generated.

        Supports both .marshal and .many().marshal Kim interfaces.  Handles errors raised
        during marshalling and automatically returns a HTTP error response.

        :param data: Data to be marshaled.
        :returns: Marshaled object according to mapper configuration
        :raises: :class:`werkzeug.exceptions.UnprocessableEntity`
        """
        try:
            if self.many:
                return self.mapper.many(raw=self.raw, **self.mapper_kwargs).marshal(
                    data, role=self.role
                )
            else:
                return self.mapper(
                    data=data,
                    obj=self.obj,
                    partial=self.partial,
                    **self.mapper_kwargs
                ).marshal(role=self.role)
        except MappingInvalid as e:
            payload = {
                "message": "Invalid or incomplete data provided.",
                "errors": e.errors
            }
            self.endpoint.return_error(error_status, payload=payload)


class KimEndpoint(Endpoint):
    """This Endpoint provides all the functionality of the normal Arrested Endpoint class
    plus a few helpers for working with Kim to avoid duplicate code.
    """

    many = False
    mapper_class = None
    serialize_role = '__default__'
    marshal_role = '__default__'
    response_handler = KimResponseHandler
    request_handler = KimRequestHandler

    def _is_marshal_request(self):
        """Return a boolean indicating wether the request is a POST PATCH or PUT request
        """
        return request.method in ['POST', 'PATCH', 'PUT']

    def is_partial(self):
        """Return a boolean indicating if this marshal request is providing partial data
        """
        return request.method == 'PATCH'

    def get_response_handler_params(self, **params):
        """Return a config object that will be used to configure the KimResponseHandler

        :returns: a dictionary of config options
        :rtype: dict
        """

        params = super(KimEndpoint, self).get_response_handler_params(**params)
        params['mapper_class'] = self.mapper_class
        params['role'] = self.serialize_role

        # After a successfull attempt to marshal an object has been made, a response
        # is generated using the RepsonseHandler.  Rather than taking the class level
        # setting for many by default, pull it from the request handler params config to
        # ensure Marshaling and Serializing are run the same way.
        if self._is_marshal_request():
            req_params = self.get_request_handler_params()
            params['many'] = req_params.get('many', self.many)
        else:
            params['many'] = self.many

        return params

    def get_request_handler_params(self, **params):
        """Return a config object that will be used to configure the KimRequestHandler

        :returns: a dictionary of config options
        :rtype: dict
        """

        params = super(KimEndpoint, self).get_request_handler_params(**params)
        params['mapper_class'] = self.mapper_class
        params['role'] = self.marshal_role
        params['many'] = False
        # when handling a PUT or PATCH request, self.obj will be set.. There might be a
        # more robust way to handle this?
        params['obj'] = getattr(self, 'obj', None)
        params['partial'] = self.is_partial()

        return params



from flask import current_app as app

from arrested.exceptions import ArrestedException

from ..mixins import (
    GetListMixin, CreateMixin, GetObjectMixin,
    PutObjectMixin, PatchObjectMixin,
    DeleteObjectMixin
)


class DBMixin(object):

    def get_query(self):
        """Return an SQLAlchemy Query object.  Users using this mixin should  override
        this method.

        The Mixins that call this method expect a Query method to be returned so
        that they can apply additional filtering automatically.
        Things like pagination middleware also require a Query object.

        Useage::

            class MyEndpoint(Endpoint, DBObjectMixin):

                def get_query(self):
                    return db.session.query(User).filter(
                        User.is_active == true()
                    )

        In the example above, Arrested will automatically apply a primary key filter to
        the query returned by :meth:`DBMixin.get_query`.

        :returns: SQLAlchemy Query object
        :raises: NotImplementedError

        .. seealso::

            :meth:`DBObjectMixin.get_objects`
            :meth:`DBListMixin.get_result`
            :meth:`DBObjectMixin.get_object`
            :meth:`DBObjectMixin.get_result`
        """
        name = self.__class__.__name__
        raise NotImplementedError('%s must implement get_query method.' % name)

    def get_db_session(self):
        """Returns the session configured against the Flask appliction instance.
        """
        return app.extensions['sqlalchemy'].db.session

    def save(self, obj):
        """Add ``obj`` to the SQLAlchemy session and commit the changes back to
        the database.

        :param obj: SQLAlchemy object being saved
        :returns: The saved object
        """
        session = self.get_db_session()
        session.add(obj)
        session.commit()

        return obj


class DBListMixin(GetListMixin, DBMixin):
    """SQLAlchemy powered list mixin that queries a database to retun a list of results
    for a given endpoint.

    Usage::

        from arrested import Resource, Endpoint
        from arrested.contrib.sql_alchemy import DBListMixin
        from arrested.contrib.kim_arrested import KimRequestHandler, KimResponseHandler

        from .models import db, Character

        characters_resource = Resource('characters', __name__, url_prefix='/characters')

        class CharactersEndpoint(Endpoint, DBListMixin):

            name = 'list'
            response_handler = KimResponseHandler
            request_handler = KimRequestHandler

            def get_query(self):

                return db.session.query(Character).order_by(
                    Character.created_at.desc()
                ).all()

        characters_resource.add_endpoint(CharactersEndpoint)

    """

    def get_result(self, query):
        """Cast the query into a scalar python type.

        :param query: SQLAlchemy Query
        :returns: A list of objects returned by the Query or an empty list
        """

        return query.all()

    def get_objects(self):
        """Implements the GetListMixin interface and calls :meth:`DBListMixin.get_query`.
        Using this mixin requires usage of a response handler capable of serializing
        SQLAlchemy query result objects.

        :returns: Typically a SQLALchemy Query result.
        :rtype: mixed

        .. seealso::
            :meth:`DBListMixin.get_query`
            :meth:`DBListMixin.get_result`
        """

        return self.get_result(self.get_query())


class DBCreateMixin(CreateMixin, DBMixin):
    """SQLAlchemy Create mixin that automatically adds new obejcts to a session object
    and commits.
    """

    def save_object(self, obj):
        """Takes a marshalled object and attempts to save it to the database using
        a SQlAlchemy session.
        """
        self.save(obj)


class DBObjectMixin(GetObjectMixin,
                    PutObjectMixin,
                    PatchObjectMixin,
                    DeleteObjectMixin,
                    DBMixin):
    """SQLAlchemy powered object mixin that queries a database to retun a single object
    typically by primary key for a given endpoint.

    Usage::

        from arrested import Resource, Endpoint
        from arrested.contrib.sql_alchemy import DBObjectMixin
        from arrested.contrib.kim_arrested import KimRequestHandler, KimResponseHandler

        from .models import db, Character

        characters_resource = Resource('characters', __name__, url_prefix='/characters')

        class CharacterObjectEndpoint(Endpoint, DBObjectMixin):

            url = '/<integer:obj_id>'
            name = 'object'
            response_handler = KimResponseHandler
            request_handler = KimRequestHandler

            def get_query(self):

                return db.session.query(Character).filter()

        characters_resource.add_endpoint(CharacterObjectEndpoint)
    """

    model = None
    url_id_param = 'obj_id'
    model_id_param = 'id'

    def get_result(self, query):
        """Cast the query into a scalar python type.

        :param query: SQLAlchemy Query
        :returns: An object returned by the Query or None
        """
        return query.one_or_none()

    def filter_by_id(self, query):
        """Apply the primary key filter to query to filter the results for a specific
        instance by id.

        The filter applied by the this method by default can be controlled using the
        url_id_param

        :param query: SQLAlchemy Query
        :returns: A SQLAlchemy Query object
        """
        if self.model is None:
            raise ArrestedException('DBObjectMixin requires a model to be set.')

        idfield = getattr(self.model, self.model_id_param, None)
        if not idfield:
            raise ArrestedException('DBObjectMixin could not find a valid Model.id.')

        return query.filter(idfield == self.kwargs[self.url_id_param])

    def get_object(self):
        """Implements the GetObjectMixin interface and calls
        :meth:`DBObjectMixin.get_query`.  Using this mixin requires usage of
        a response handler capable of serializing SQLAlchemy query result objects.

        :returns: Typically a SQLALchemy Query result.
        :rtype: mixed

        .. seealso::
            :meth:`DBObjectMixin.get_query`
            :meth:`DBObjectMixin.filter_by_id`
            :meth:`DBObjectMixin.get_result`
        """

        query = self.get_query()
        query = self.filter_by_id(query)
        return self.get_result(query)

    def update_object(self, obj):
        """Commits changes to an instance back to the database by
        calling :meth:`.DBMixin.save` on the provided object.
        """
        return self.save(obj)

    def patch_object(self, obj):
        """Commits changes to an instance back to the database by
        calling :meth:`.DBMixin.save` on the provided object.
        """
        return self.save(obj)

    def delete_object(self, obj):
        """Deletes an object from the session by calling session.delete and then commits
        the changes to the database.

        :param obj: The SQLAlchemy instance being deleted
        :returns: None
        """
        session = self.get_db_session()
        session.delete(obj)
        session.commit()



import pytest

from kim import Mapper, field

from werkzeug.exceptions import BadRequest, UnprocessableEntity
from mock import MagicMock

from arrested import PutObjectMixin, PatchObjectMixin
from arrested.contrib.kim_arrested import (
    KimHandler, KimResponseHandler,
    KimRequestHandler,
    KimEndpoint
)

from tests.endpoints import CharactersEndpoint


class MyObject(object):

    def __init__(self, id=None, name=''):

        self.id = id
        self.name = name

    def __eq__(self, other):

        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False


class MyMapper(Mapper):

    __type__ = MyObject

    id = field.Integer()
    name = field.String()

    __roles__ = {
        'name_only': ['name']
    }


def test_kim_invalid_mapper_class():
    with pytest.raises(Exception):
        endpoint = CharactersEndpoint()
        KimHandler(endpoint, mapper_class=None).mapper


def test_kim_handler_get_mapper():

    endpoint = CharactersEndpoint()
    handler = KimHandler(endpoint, mapper_class=MyMapper)
    assert handler.mapper == MyMapper


def test_kim_handler_sets_opts():

    endpoint = CharactersEndpoint()
    handler = KimHandler(
        endpoint,
        mapper_class=MyMapper,
        mapper_kwargs={'foo': 'bar'},
        role='test',
        many=True,
        raw=True)
    assert handler.mapper_class == MyMapper
    assert handler.mapper_kwargs == {'foo': 'bar'}
    assert handler.role == 'test'
    assert handler.many is True
    assert handler.raw is True


def test_kim_handler_default_opts():

    endpoint = CharactersEndpoint()
    handler = KimHandler(endpoint)
    assert handler.mapper_class is None
    assert handler.mapper_kwargs == {}
    assert handler.role == '__default__'
    assert handler.many is False
    assert handler.raw is False


def test_kim_response_handler_with_many():

    data = [MyObject(id=1, name='test 1'), MyObject(id=2, name='test 2')]
    endpoint = CharactersEndpoint()
    handler = KimResponseHandler(endpoint, mapper_class=MyMapper, many=True)

    resp = handler.handle(data)
    assert resp == [{'id': 1, 'name': 'test 1'}, {'id': 2, 'name': 'test 2'}]


def test_kim_response_handler_with_many_and_role():

    data = [MyObject(id=1, name='test 1'), MyObject(id=2, name='test 2')]
    endpoint = CharactersEndpoint()
    handler = KimResponseHandler(
        endpoint, mapper_class=MyMapper, many=True, role='name_only')

    resp = handler.handle(data)
    assert resp == [{'name': 'test 1'}, {'name': 'test 2'}]


def test_kim_response_handler():

    data = MyObject(id=1, name='test 1')
    endpoint = CharactersEndpoint()
    handler = KimResponseHandler(endpoint, mapper_class=MyMapper, many=False)

    resp = handler.handle(data)
    assert resp == {'id': 1, 'name': 'test 1'}


def test_kim_response_handler_role():

    data = MyObject(id=1, name='test 1')
    endpoint = CharactersEndpoint()
    handler = KimResponseHandler(endpoint, mapper_class=MyMapper, role='name_only')

    resp = handler.handle(data)
    assert resp == {'name': 'test 1'}


def test_kim_response_handler_mapper_kwargs():

    mock_mapper = MagicMock(spec=MyMapper)
    data = MyObject(id=1, name='test 1')
    endpoint = CharactersEndpoint()
    handler = KimResponseHandler(
        endpoint, mapper_class=mock_mapper, mapper_kwargs={'test': 1})

    handler.handle(data)
    assert mock_mapper.call_count == 1
    call_args, call_kwargs = mock_mapper.call_args
    assert call_kwargs == {'obj': data, 'raw': False, 'test': 1}


def test_kim_request_handler():

    data = {'name': 'test 1', 'id': 1}
    endpoint = CharactersEndpoint()
    handler = KimRequestHandler(endpoint, mapper_class=MyMapper)

    resp = handler.handle(data)
    assert resp == MyObject(id=1, name='test 1')


def test_kim_request_handler_with_many():

    data = [{'name': 'test 1', 'id': 1}, {'name': 'test 2', 'id': 2}]
    endpoint = CharactersEndpoint()
    handler = KimRequestHandler(endpoint, mapper_class=MyMapper, many=True)

    resp = handler.handle(data)
    assert resp == [MyObject(id=1, name='test 1'), MyObject(id=2, name='test 2')]


def test_kim_request_handler_with_errors():

    data = [{'id': 1}, {'name': 'test 2'}]
    endpoint = CharactersEndpoint()
    handler = KimRequestHandler(endpoint, mapper_class=MyMapper, many=True)

    with pytest.raises(UnprocessableEntity):

        handler.handle(data)


def test_kim_request_handler_with_errors_resp():

    data = [{'id': 1}, {'name': 'test 2'}]
    endpoint = CharactersEndpoint()
    handler = KimRequestHandler(endpoint, mapper_class=MyMapper, many=True)

    try:
        handler.handle(data)
        assert False, 'Did not raise UnprocessableEntity'
    except UnprocessableEntity as resp:
        assert resp.response.data == \
            b'{"message": "Invalid or incomplete data provided.", "errors": {"name": "This is a required field"}}'


def test_kim_request_handler_with_custom_error_status():

    data = [{'id': 1}, {'name': 'test 2'}]
    endpoint = CharactersEndpoint()
    handler = KimRequestHandler(endpoint, mapper_class=MyMapper, many=True)

    try:
        handler.handle(data, error_status=400)
        assert False, 'Did not raise UnprocessableEntity'
    except BadRequest as resp:
        assert resp.response.data == \
            b'{"message": "Invalid or incomplete data provided.", "errors": {"name": "This is a required field"}}'


def test_kim_endpoint_is_marshal_request_false(app):

    with app.test_request_context('/test', method='GET'):
        endpoint = KimEndpoint()
        assert endpoint._is_marshal_request() == False


def test_kim_endpoint_is_marshal_request_true(app):

    def _test(method):
        with app.test_request_context('/test', method=method):
            endpoint = KimEndpoint()
            assert endpoint._is_marshal_request() == True

    [_test(method) for method in ['POST', 'PUT', 'PATCH']]


def test_kim_endpoint_get_response_handler_params(app):

    class MyEndpoint(KimEndpoint):
        mapper_class = MyMapper
        many = True
        serialize_role = 'test'

    with app.test_request_context('/test'):
        endpoint = MyEndpoint()
        params = endpoint.get_response_handler_params()
        exp = {
            'mapper_class': MyMapper,
            'many': True,
            'role': 'test'
        }
        assert params == exp


def test_kim_endpoint_get_request_handler_params(app):

    class MyEndpoint(KimEndpoint):
        mapper_class = MyMapper
        many = True
        marshal_role = 'test'

    with app.test_request_context('/test', method="POST"):
        endpoint = MyEndpoint()
        params = endpoint.get_request_handler_params()
        exp = {
            'mapper_class': MyMapper,
            'many': False,
            'role': 'test',
            'obj': None,
            'partial': False,
        }
        assert params == exp


def test_kim_endpoint_get_request_handler_params_update_existing(app):

    class MyEndpoint(KimEndpoint, PutObjectMixin):
        mapper_class = MyMapper
        many = True
        marshal_role = 'test'

        def get_object(self):

            return {'foo': 'bar'}

    with app.test_request_context('/test', method="PUT"):
        endpoint = MyEndpoint()
        params = endpoint.get_request_handler_params()
        exp = {
            'mapper_class': MyMapper,
            'many': False,
            'role': 'test',
            'obj': {'foo': 'bar'},
            'partial': False,
        }
        assert params == exp


def test_kim_endpoint_get_request_handler_partial_update(app):

    class MyEndpoint(KimEndpoint, PatchObjectMixin):
        mapper_class = MyMapper
        many = True
        marshal_role = 'test'

        def get_object(self):

            return {'foo': 'bar'}

    with app.test_request_context('/test', method="PATCH"):
        endpoint = MyEndpoint()
        params = endpoint.get_request_handler_params()
        exp = {
            'mapper_class': MyMapper,
            'many': False,
            'role': 'test',
            'obj': {'foo': 'bar'},
            'partial': True,
        }
        assert params == exp


def test_kim_endpoint_get_response_handler_params_sets_many_from_request_params(app):

    class MyEndpoint(KimEndpoint):
        mapper_class = MyMapper
        many = False
        marshal_role = 'test'

        def get_request_handler_params(self, **params):

            params = super(MyEndpoint, self).get_request_handler_params(**params)
            params['many'] = True

            return params

    with app.test_request_context('/test', method="POST"):
        endpoint = MyEndpoint()
        params = endpoint.get_response_handler_params()
        exp = {
            'mapper_class': MyMapper,
            'many': True,
            'role': '__default__'
        }
        assert params == exp



import os
import sys
from distutils.sysconfig import get_python_lib

from setuptools import find_packages, setup

setup(
    name='arrested',
    version='0.1.1',
    author='Mikey Waites',
    author_email='mike@oldstlabs.com',
    url='https://github.com/mikeywaites/flask-arrested',
    description=('A framework for rapidly building REST APIs in Flask.'),
    license='MIT',
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    install_requires=['flask'],
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Flask',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
